/*
 * Copyright 2004-2009 the Seasar Foundation and the Others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied. See the License for the specific language
 * governing permissions and limitations under the License.
 */
package org.seasar.doma.internal.jdbc.query;

import static org.seasar.doma.internal.util.AssertionUtil.*;

import java.sql.Statement;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.seasar.doma.internal.jdbc.entity.EntityPropertyType;
import org.seasar.doma.internal.jdbc.entity.EntityType;
import org.seasar.doma.internal.jdbc.entity.EntityTypeFactory;
import org.seasar.doma.internal.jdbc.entity.GeneratedIdPropertyType;
import org.seasar.doma.internal.jdbc.sql.PreparedSql;
import org.seasar.doma.internal.jdbc.sql.PreparedSqlBuilder;
import org.seasar.doma.internal.message.DomaMessageCode;
import org.seasar.doma.jdbc.JdbcException;
import org.seasar.doma.jdbc.id.IdGenerationConfig;

/**
 * @author taedium
 * 
 */
public class AutoBatchInsertQuery<E> extends AutoBatchModifyQuery<E> implements
        BatchInsertQuery {

    protected GeneratedIdPropertyType<?> generatedIdPropertyType;

    protected final List<GeneratedIdPropertyType<?>> generatedIdPropertyTypes = new ArrayList<GeneratedIdPropertyType<?>>();

    protected IdGenerationConfig idGenerationConfig;

    protected boolean batchSupported = true;

    public AutoBatchInsertQuery(EntityTypeFactory<E> entityTypeFactory) {
        super(entityTypeFactory);
    }

    public void prepare() {
        assertNotNull(config);
        Iterator<EntityType<E>> it = entityTypes.iterator();
        if (it.hasNext()) {
            executable = true;
            executionSkipCause = null;
            entityType = it.next();
            entityType.preInsert();
            prepareTableAndColumnNames();
            prepareIdAndVersionProperties();
            prepareOptions();
            prepareTargetProperties();
            prepareVersionValue();
            prepareSql();
        } else {
            return;
        }
        while (it.hasNext()) {
            idProperties.clear();
            generatedIdPropertyType = null;
            versionPropertyType = null;
            targetProperties.clear();
            this.entityType = it.next();
            entityType.preInsert();
            prepareIdAndVersionProperties();
            prepareTargetProperties();
            prepareVersionValue();
            prepareSql();
        }
        assertEquals(entityTypes.size(), sqls.size());
    }

    @Override
    protected void prepareIdAndVersionProperties() {
        super.prepareIdAndVersionProperties();
        generatedIdPropertyType = entityType.getGeneratedIdPropertyType();
        generatedIdPropertyTypes.add(generatedIdPropertyType);
        if (generatedIdPropertyType != null) {
            if (idGenerationConfig == null) {
                String idColumnName = columnNameMap.get(generatedIdPropertyType
                        .getName());
                idGenerationConfig = new IdGenerationConfig(config, entityType,
                        tableName, idColumnName);
                generatedIdPropertyType
                        .validateGenerationStrategy(idGenerationConfig);
                autoGeneratedKeysSupported = generatedIdPropertyType
                        .isAutoGeneratedKeysSupported(idGenerationConfig);
                batchSupported = generatedIdPropertyType
                        .isBatchSupported(idGenerationConfig);
            }
            generatedIdPropertyType.preInsert(idGenerationConfig);
        }
    }

    protected void prepareTargetProperties() {
        for (EntityPropertyType<?> p : entityType.getEntityPropertyTypes()) {
            if (!p.isInsertable()) {
                continue;
            }
            if (p.isId()) {
                if (p != generatedIdPropertyType
                        || generatedIdPropertyType
                                .isIncluded(idGenerationConfig)) {
                    targetProperties.add(p);
                }
                if (generatedIdPropertyType == null
                        && p.getWrapper().get() == null) {
                    throw new JdbcException(DomaMessageCode.DOMA2020,
                            entityType.getName(), p.getName());
                }
                continue;
            }
            if (!isTargetPropertyName(p.getName())) {
                continue;
            }
            targetProperties.add(p);
        }
    }

    protected void prepareVersionValue() {
        if (versionPropertyType != null) {
            versionPropertyType.setIfNecessary(1);
        }
    }

    protected void prepareSql() {
        PreparedSqlBuilder builder = new PreparedSqlBuilder(config);
        builder.appendSql("insert into ");
        builder.appendSql(tableName);
        builder.appendSql(" (");
        for (EntityPropertyType<?> p : targetProperties) {
            builder.appendSql(columnNameMap.get(p.getName()));
            builder.appendSql(", ");
        }
        builder.cutBackSql(2);
        builder.appendSql(") values (");
        for (EntityPropertyType<?> p : targetProperties) {
            builder.appendWrapper(p.getWrapper());
            builder.appendSql(", ");
        }
        builder.cutBackSql(2);
        builder.appendSql(")");
        PreparedSql sql = builder.build();
        sqls.add(sql);
    }

    @Override
    public boolean isBatchSupported() {
        return batchSupported;
    }

    @Override
    public void generateId(Statement statement, int index) {
        GeneratedIdPropertyType<?> generatedIdProperty = generatedIdPropertyTypes
                .get(index);
        generatedIdProperty.postInsert(idGenerationConfig, statement);
    }

}
